# py2exe.py - Functionality for performing py2exe builds.
#
# Copyright 2019 Gregory Szorc <gregory.szorc@gmail.com>
# Copyright 2022 Matt Harbison <mharbison72@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

# no-check-code because Python 3 native.

import distutils.sysconfig
import os
import pathlib
import shutil
import subprocess
import sys

from .downloads import download_entry
from .util import (
    extract_tar_to_directory,
    extract_zip_to_directory,
    process_install_rules,
    python_exe_info,
    SourceDirs,
)


STAGING_RULES = [
    ('contrib/bash_completion', 'contrib/'),
    ('contrib/hgk', 'contrib/hgk.tcl'),
    ('contrib/hgweb.fcgi', 'contrib/'),
    ('contrib/hgweb.wsgi', 'contrib/'),
    ('contrib/logo-droplets.svg', 'contrib/'),
    ('contrib/mercurial.el', 'contrib/'),
    ('contrib/mq.el', 'contrib/'),
    ('contrib/tcsh_completion', 'contrib/'),
    ('contrib/tcsh_completion_build.sh', 'contrib/'),
    ('contrib/vim/*', 'contrib/vim/'),
    ('contrib/win32/postinstall.txt', 'ReleaseNotes.txt'),
    ('contrib/win32/ReadMe.html', 'ReadMe.html'),
    ('contrib/xml.rnc', 'contrib/'),
    ('contrib/zsh_completion', 'contrib/'),
    ('dist/hg.exe', './'),
    ('dist/lib/*.dll', 'lib/'),
    ('dist/lib/*.pyd', 'lib/'),
    ('dist/lib/library.zip', 'lib/'),
    ('dist/python*.dll', './'),
    ('doc/*.html', 'doc/'),
    ('doc/style.css', 'doc/'),
    ('mercurial/helptext/**/*.txt', 'helptext/'),
    ('mercurial/defaultrc/*.rc', 'defaultrc/'),
    ('mercurial/locale/**/*', 'locale/'),
    ('mercurial/templates/**/*', 'templates/'),
    ('COPYING', 'Copying.txt'),
]

# List of paths to exclude from the staging area.
STAGING_EXCLUDES = [
    'doc/hg-ssh.8.html',
]


def build_py2exe(
    source_dirs: SourceDirs,
    build_dir: pathlib.Path,
    python_exe: pathlib.Path,
    build_name: str,
    venv_requirements_txt: pathlib.Path,
    extra_packages=None,
    extra_excludes=None,
    extra_dll_excludes=None,
    extra_packages_script=None,
    extra_includes=None,
):
    """Build TortoiseHg with py2exe.

    Build files will be placed in ``build_dir``.
    """

    env = dict(os.environ)
    env["HGPLAIN"] = "1"
    env["HGRCPATH"] = ""

    site_packages_path = distutils.sysconfig.get_python_lib()

    included_packages = ['mercurial']

    # make sure that packages globally installed with e.g. easy_install
    # don't override packages we want to bundle and build ourselves
    err = set()
    for m in included_packages:
        try:
            mod = __import__(m)
        except ImportError:
            continue
        fm = mod.__file__
        if fm.lower().startswith(site_packages_path.lower()):
            print(
                "Error: '%s' overrides included package '%s'"
                % (os.path.dirname(fm), m)
            )
            err.add(m)
    if err:
        print(
            "(uninstall or hide these installed packages: %s)" % ', '.join(err)
        )
        sys.exit(1)

    # Locate the HTML Help Workshop installation
    if "PROGRAMW6432" in env:
        programfiles_x86 = env.get("PROGRAMFILES(X86)")
    else:
        programfiles_x86 = env.get("PROGRAMFILES")

    if programfiles_x86:
        env['PATH'] = "%s%s%s" % (
            env['PATH'],
            os.pathsep,
            os.path.join(programfiles_x86, 'HTML Help Workshop')
        )

    # If HTML Help Workshop needs to be installed, it can be found here:
    # https://download.microsoft.com/download/OfficeXPProf/Install/4.71.1015.0/W98NT42KMe/EN-US/HTMLHELP.EXE
    hhc = shutil.which("hhc.exe", path=env['PATH'])
    if not hhc:
        raise Exception("Unable to find HTML Help Workshop")

    # The thg/doc/build.bat task fails unless this is set explicitly.
    env['hhc_compiler'] = hhc

    py_info = python_exe_info(python_exe)

    vc_x64 = py_info['arch'] == '64bit'

    build_dir.mkdir(exist_ok=True)

    gettext_pkg, gettext_entry = download_entry('gettext', build_dir)
    gettext_dep_pkg = download_entry('gettext-dep', build_dir)[0]

    venv_path = build_dir / (
        'venv-%s-%s' % (build_name, 'x64' if vc_x64 else 'x86')
    )

    gettext_root = build_dir / ('gettext-win-%s' % gettext_entry['version'])

    if not gettext_root.exists():
        extract_zip_to_directory(gettext_pkg, gettext_root)
        extract_zip_to_directory(gettext_dep_pkg, gettext_root)

    if not venv_path.exists():
        print('creating virtualenv with dependencies')
        subprocess.run(
            [str(python_exe), "-m", "venv", str(venv_path)], check=True
        )

    venv_python = venv_path / 'Scripts' / 'python.exe'
    venv_pip = venv_path / 'Scripts' / 'pip.exe'

    subprocess.run(
        [str(venv_pip), 'install', '-r', str(venv_requirements_txt)], check=True
    )

    if extra_packages_script:
        more_packages = set(
            subprocess.check_output(extra_packages_script, cwd=build_dir)
            .split(b'\0')[-1]
            .strip()
            .decode('utf-8')
            .splitlines()
        )
        if more_packages:
            if not extra_packages:
                extra_packages = more_packages
            else:
                extra_packages |= more_packages

    if extra_packages:
        env['HG_PY2EXE_EXTRA_PACKAGES'] = ' '.join(sorted(extra_packages))
        hgext3rd_extras = sorted(
            e for e in extra_packages if e.startswith('hgext3rd.')
        )
        if hgext3rd_extras:
            env['HG_PY2EXE_EXTRA_INSTALL_PACKAGES'] = ' '.join(hgext3rd_extras)
    if extra_includes:
        env['HG_PY2EXE_EXTRA_INCLUDES'] = ' '.join(sorted(extra_includes))
    if extra_excludes:
        env['HG_PY2EXE_EXTRA_EXCLUDES'] = ' '.join(sorted(extra_excludes))
    if extra_dll_excludes:
        env['HG_PY2EXE_EXTRA_DLL_EXCLUDES'] = ' '.join(
            sorted(extra_dll_excludes)
        )

    # Register location of msgfmt and other binaries.
    env['PATH'] = '%s%s%s' % (
        env['PATH'],
        os.pathsep,
        str(gettext_root / 'bin'),
    )

    print('building Mercurial')
    hg_dir = source_dirs.hg

    # Clear files added by previous installer runs
    subprocess.run(
        ["hg.exe", "--config", "extensions.purge=", "purge", "--all"],
        cwd=str(hg_dir),
        env=env,
        check=True,
    )

    oldpath = env.get('PYTHONPATH', '').split(os.pathsep)
    path = [str(hg_dir), str(source_dirs.evolve)]
    path.extend(oldpath)
    env['PYTHONPATH'] = os.pathsep.join(path)

    subprocess.run(
        [str(venv_python), 'setup.py', '--version'],
        cwd=str(hg_dir),
        env=env,
        check=True,
    )
    subprocess.run(
        [str(venv_python), 'setup.py', 'build_py', '-c', '-d', '.', 'build_mo'],
        cwd=str(hg_dir),
        env=env,
        check=True,
    )
    subprocess.run(
        [str(venv_python), 'setup.py', 'build_ext', '-i'],
        cwd=str(hg_dir),
        env=env,
        check=True,
    )
    subprocess.run(
        [str(venv_python), 'setup.py', 'build_hgextindex'],
        cwd=str(hg_dir),
        env=env,
        check=True,
    )
    # TODO: the zstd entry isn't being added to doc/hg.1.html for some reason on
    #  some systems.  It would appear that cext isn't loading.  (shell=True
    #  doesn't help)
    subprocess.run(
        # Use python_exe instead of venv_python so it has access to the docutils
        # module in the bootstrap venv.
        [str(python_exe), 'setup.py', 'build_doc', '--html'],
        cwd=str(hg_dir),
        env=env,
        check=True,
    )

    print('building TortoiseHg')
    thg_dir = source_dirs.thg

    if thg_dir.exists():
        shutil.rmtree(thg_dir)
    thg_dir.mkdir()

    subprocess.run(
        ['hg.exe', 'archive', '-r', 'wdir()', str(thg_dir)],
        cwd=str(source_dirs.original),
        env=env,
        check=True,
    )

    subprocess.run(
        [str(venv_python), 'setup.py', '--version'],
        cwd=str(thg_dir),
        env=env,
        check=True,
    )
    subprocess.run(  # Build locales
        [str(venv_python), 'setup.py', 'build_mo'],
        cwd=str(thg_dir),
        env=env,
        check=True,
    )
    subprocess.run(  # Build cmenu translation registry files
        [str(venv_python), 'reggen.py'],
        cwd=str(thg_dir / "win32"),
        env=env,
        check=True,
    )

    # Use build helpers from win32
    dest_config_py = thg_dir / 'tortoisehg' / 'util' / 'config.py'
    shutil.copyfile(thg_dir / 'win32' / 'config.py', dest_config_py)

    with open(thg_dir / "setup.cfg", "w") as fp:
        fp.write('''
[py2exe]
includes = PyQt5.sip, PyQt5.QtPrintSupport, PyQt5.QtSvg, PyQt5.QtXml,
           mercurial_keyring, pygit2

packages = ctypes, curses, email, encodings, iniparse, json, keyring,
           keyring.backends, nntplib, pygments, sspi, sqlite3, six,

           hgext,
           hgext.convert, hgext.fastannotate, hgext.fsmonitor,
           hgext.fsmonitor.pywatchman, hgext.git, hgext.highlight,
           hgext.hooklib, hgext.infinitepush, hgext.largefiles,
           hgext.lfs, hgext.narrow, hgext.remotefilelog, hgext.zeroconf,
           mercurial.cext, mercurial.pure, mercurial.defaultrc,

           hgext3rd, hgext3rd.evolve, hgext3rd.topic,

           tortoisehg.hgqt, tortoisehg.util
'''.strip())

    # py2exe has trouble finding hgext3rd packages unless they're dumped into
    # the build directory.
    evolve_hgext3rd_dir = source_dirs.evolve / 'hgext3rd'
    shutil.copytree(evolve_hgext3rd_dir / 'evolve', thg_dir / 'hgext3rd' / 'evolve')
    shutil.copytree(evolve_hgext3rd_dir / 'topic', thg_dir / 'hgext3rd' / 'topic')
    shutil.rmtree(thg_dir / 'hgext3rd' / 'evolve' / 'hack')

    # Note: hgextindex was regenerated here in the winbuild script.

    subprocess.run(  # Build docs
        ['build', 'chm'],
        shell=True,
        cwd=str(thg_dir / "doc"),
        env=env,
        check=True,
    )

    env['MERCURIAL_PATH'] = str(hg_dir / "mercurial")
    env['HGEXT_PATH'] = str(hg_dir / "hgext")

    # Put pyrcc5.exe installed by pip on PATH
    env['PATH'] = "%s;%s" % (venv_python.parent, env['PATH'])

    # The py2exe target generates these in the same location for both py2 and
    # py3 builds.  But for some reason, the generated *_rc.py files aren't put
    # into library.zip as part of the py3 build process.  So generate the files
    # early and copy them to the source tree where they will be picked up.
    subprocess.run(  # Build the resource files
        [str(venv_python), 'setup.py', 'build_qrc'],
        cwd=str(thg_dir),
        env=env,
        check=True,
    )
    for f in ("icons_rc.py", "translations_rc.py"):
        shutil.copyfile(
            thg_dir / "build" / "lib" / "tortoisehg" / "hgqt" / f,
            thg_dir / "tortoisehg" / "hgqt" / f
        )

    subprocess.run(  # Build the executables
        [str(venv_python), 'setup.py', 'py2exe', '-b3'],
        cwd=str(thg_dir),
        env=env,
        check=True,
    )

    dist_dir = thg_dir / "dist"

    # For some reason, py2exe doesn't copy the -x64 named files over to dist/ at
    # all.  Both the files with and without -x64 are 64-bit in the 64-bit build.
    # The non -x64 named files get copied to dist/lib by py2exe for both 32 and
    # 64-bit builds.  In order for the update check in the aboutbox to work,
    # the -x64 named files need to be _next_ to lib/ in the 64-bit installation,
    # and the non -x64 named files next to lib/ 32-bit installation.  The non
    # -x64 named files are required in the 64-bit build, or the app fails to
    # launch.
    #
    # The check for this working properly from ``hg debugshell`` is:
    #
    # >>> from PyQt5 import QtNetwork
    # >>> print(QtNetwork.QSslSocket.sslLibraryBuildVersionString())
    # OpenSSL 1.1.1g  21 Apr 2020
    # >>> print(QtNetwork.QSslSocket.sslLibraryVersionString())
    # OpenSSL 1.1.1n  15 Mar 2022
    #
    # When not properly positioned, the output of the last line is empty.
    if vc_x64:
        path_script = """
import os, sys
join = os.path.join
print(join(sys.exec_prefix, 'lib', 'site-packages', 'PyQt5', 'Qt5', 'bin'))
""".strip()

        binpath = os.fsdecode(subprocess.check_output(
            [str(venv_python), '-c', path_script],
            cwd=str(thg_dir),
            env=env,
        ).strip())

        # For some reason, these don't load if installed in the lib directory.
        for f in ("libcrypto-1_1-x64.dll", "libssl-1_1-x64.dll"):
            shutil.copy(os.path.join(binpath, f), dist_dir)
    else:
        # For some reason, these don't load if installed in the lib directory.
        # If it is merely copied next to lib/, the presence of two files with
        # the same name confuses WiX (unless a File Id is explicitly provided),
        # so remove it from lib/.
        for f in ("libcrypto-1_1.dll", "libssl-1_1.dll"):
            shutil.move(dist_dir / "lib" / f, dist_dir)


def stage_install(
    source_dir: pathlib.Path, staging_dir: pathlib.Path, lower_case=False
):
    """Copy all files to be installed to a directory.

    This allows packaging to simply walk a directory tree to find source
    files.
    """
    if lower_case:
        rules = []
        for source, dest in STAGING_RULES:
            # Only lower directory names.
            if '/' in dest:
                parent, leaf = dest.rsplit('/', 1)
                dest = '%s/%s' % (parent.lower(), leaf)
            rules.append((source, dest))
    else:
        rules = STAGING_RULES

    process_install_rules(rules, source_dir, staging_dir)

    # Write out a default editor.rc file to configure notepad as the
    # default editor.
    with (staging_dir / 'defaultrc' / 'editor.rc').open(
        'w', encoding='utf-8'
    ) as fh:
        fh.write('[ui]\neditor = notepad\n')

    # Purge any files we don't want to be there.
    for f in STAGING_EXCLUDES:
        p = staging_dir / f
        if p.exists():
            print('removing %s' % p)
            p.unlink()
